#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2010, 2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
from os.path import abspath, dirname, join
sys.path.insert(0, join(abspath(dirname(__file__)), 'lib'))
from skoolparser import SkoolParser, get_address, parse_int

BLOCKS = 'b'
BLOCK_TITLES = 't'
BLOCK_DESC = 'd'
REGISTERS = 'r'
BLOCK_COMMENTS = 'm'
SUBBLOCKS = 's'
COMMENTS = 'c'

class FakeInstruction:
    def __init__(self, address, comment):
        self.address = address
        self.comment = comment

def write_line(text):
    sys.stdout.write('%s\n' % text)

def get_instruction_ctl(op):
    if op.startswith('DEFB'):
        return 'b'
    if op.startswith('DEFW'):
        return 'w'
    if op.startswith('DEFM'):
        return 't'
    if op.startswith('DEFS'):
        return 'z'
    return 'c'

def get_defb_length(op):
    return op.count(',') + 1

def get_defw_length(op):
    return 2 * (op.count(',') + 1)

def get_defm_length(op):
    start = op.index('"')
    return op.index('"', start + 1) - start - 1

def addr_str(address):
    return '$%04X' % int(address) if WRITE_HEX else str(address)

def num_str(num):
    if not WRITE_HEX:
        return str(num)
    integer = int(num)
    if integer < 10:
        fmt = '%i'
    elif integer < 256:
        fmt = '$%02X'
    else:
        fmt = '$%04X'
    return fmt % integer

def need_sub_block(ctl, entry_ctl, comment, index, instructions):
    # Return whether a ctl line needs to be written for a sub-block
    if ctl == entry_ctl and not comment:
        # Don't write a ctl line for a commentless sub-block of a code block
        if ctl == 'c':
            return False
        if index == 0 and len(instructions) == 1:
            # Don't write a ctl line for a DEF{B,M,W} statement if it's 1 byte
            # or 1 word, commentless, the first instruction in the section, the
            # sole instruction in its sub-block, and of the same type as the
            # containing block (i.e. 'b', 't' or 'w')
            if ctl == 'b' and get_defb_length(instructions[0].operation) == 1:
                return False
            if ctl == 't' and get_defm_length(instructions[0].operation) == 1:
                return False
            if ctl == 'w' and get_defw_length(instructions[0].operation) == 2:
                return False
            # Don't write a ctl line for a DEFS statement if it's commentless,
            # the first instruction in the section, the sole instruction in its
            # sub-block, and inside a 'z' block
            if ctl == 'z':
                return False
    return True

def get_sub_blocks(instructions):
    # Split a block of instructions into sub-blocks by comment rowspan and/or
    # instruction type
    sub_blocks = []
    i = 0
    prev_ctl = ''
    while i < len(instructions):
        instruction = instructions[i]
        comment = instruction.comment
        ctl = get_instruction_ctl(instruction.operation)
        if comment and (comment.rowspan > 1 or comment.text):
            inst_ctls = set()
            for inst in instructions[i:i + comment.rowspan]:
                inst_ctls.add(get_instruction_ctl(inst.operation))
            if len(inst_ctls) > 1:
                # We've found a set of two or more instructions of various
                # types with a single comment, so add a commented 'M' sub-block
                # and commentless sub-blocks for the instructions
                sub_blocks.append(('M', [FakeInstruction(instruction.address, instruction.comment)]))
                instruction.comment = None
                sub_blocks += get_sub_blocks(instructions[i:i + comment.rowspan])
            else:
                # We've found a set of one or more instructions of the same
                # type with a comment, so add a new sub-block
                sub_blocks.append((ctl, instructions[i:i + comment.rowspan]))
            prev_ctl = ''
        elif ctl == prev_ctl:
            # This instruction is commentless and is of the same type as the
            # previous instruction (which was also commentless), so add it to
            # the current sub-block
            sub_blocks[-1][1].append(instruction)
        else:
            # This instruction is commentless but of a different type from the
            # previous instruction, so start a new sub-block
            sub_blocks.append((ctl, [instruction]))
            prev_ctl = ctl
        i += comment.rowspan if comment else 1
    return sub_blocks

def write_sub_block(ctl, entry_ctl, comment, index, instructions, length):
    sub_block_ctl = ' ' if ctl == entry_ctl else ctl.upper()
    address = addr_str(instructions[0].address)
    lengths = ''

    if ctl in 'btwz':
        # Find the byte lengths of each line in a 'B', 'T', 'W' or 'Z'
        # sub-block
        stmt_lengths = []
        for stmt in instructions:
            operation = stmt.operation
            if ctl == 'b':
                stmt_lengths.append(get_defb_length(operation))
            elif ctl == 't':
                stmt_lengths.append(get_defm_length(operation))
            elif ctl == 'w':
                stmt_lengths.append(get_defw_length(operation))
            else:
                stmt_lengths.append(parse_int(get_address(operation[4:])))
        num_bytes = sum(stmt_lengths)
        while len(stmt_lengths) > 1 and stmt_lengths[-1] == stmt_lengths[-2]:
            stmt_lengths.pop()
        # Don't write the statement lengths in a ctl line for a lone DEFB
        # statement of length 1, a lone DEFW statement of length 2, or a lone
        # DEFS statement
        if not (len(instructions) == 1 and ((ctl == 'b' and num_bytes == 1) or (ctl == 'w' and num_bytes == 2) or ctl == 'z')):
            length = num_bytes
            lengths = ',' + ','.join([num_str(d) for d in stmt_lengths])

    if length:
        write_line(('%s %s,%s%s %s' % (sub_block_ctl, address, num_str(length), lengths, comment)).rstrip())
    else:
        write_line(('%s %s %s' % (sub_block_ctl, address, comment)).rstrip())

def write_body(entry, elements):
    entry_ctl = 'b' if entry.ctl in 'gu' else entry.ctl
    first_instruction = entry.instructions[0]
    if entry_ctl == 'i' and not first_instruction.operation:
        # Don't write any sub-blocks for an empty 'i' entry
        return

    # Split the entry into sections separated by mid-routine comments
    sections = [('', [first_instruction])]
    for instruction in entry.instructions[1:]:
        mrc = instruction.get_mid_routine_comment()
        if mrc:
            sections.append((mrc, [instruction]))
        else:
            sections[-1][1].append(instruction)

    for k, (mrc, instructions) in enumerate(sections):
        if BLOCK_COMMENTS in elements and mrc:
            address = instructions[0].address
            write_line('D %s %s' % (addr_str(address), mrc))

        if SUBBLOCKS in elements:
            sub_blocks = get_sub_blocks(instructions)
            for j, (ctl, instructions) in enumerate(sub_blocks):
                if ctl != 'M' or COMMENTS in elements:
                    length = None
                    offset = instructions[0].comment.rowspan + 1 if ctl == 'M' else 1
                    if j + offset < len(sub_blocks):
                        length = int(sub_blocks[j + offset][1][0].address) - int(instructions[0].address)
                    elif k + 1 < len(sections):
                        length = int(sections[k + 1][1][0].address) - int(instructions[0].address)
                    comment = ''
                    if instructions[0].comment and COMMENTS in elements:
                        comment = instructions[0].comment.text
                    if need_sub_block(ctl, entry_ctl, comment, j, instructions):
                        write_sub_block(ctl, entry_ctl, comment, j, instructions, length)

def write_entry(entry, elements):
    address = addr_str(entry.address)
    if BLOCKS in elements:
        if BLOCK_TITLES in elements:
            write_line(('%s %s %s' % (entry.ctl, address, entry.description)).rstrip())
        else:
            write_line('%s %s' % (entry.ctl, address))
    if BLOCK_DESC in elements:
        if entry.details:
            write_line('D %s %s' % (address, entry.details))
    if REGISTERS in elements:
        for reg in entry.registers:
            write_line('R %s %s %s' % (address, reg.name, reg.contents))
    write_body(entry, elements)
    if BLOCK_COMMENTS in elements and entry.end_comment:
        write_line('E %s %s' % (address, entry.end_comment))

def write_ctls(entries, elements):
    for entry in entries:
        write_entry(entry, elements)

def parse_options(options):
    elements = BLOCKS + BLOCK_TITLES + BLOCK_DESC + REGISTERS + BLOCK_COMMENTS + SUBBLOCKS + COMMENTS
    write_hex = False
    i = 0
    while i < len(options):
        option = options[i]
        if option[:2] == '-w':
            elements = option[2:]
        elif option == '-h':
            write_hex = True
        i += 1
    return elements, write_hex

################################################################################
# Begin
################################################################################
if len(sys.argv) < 2:
    sys.stderr.write('Usage: %s [options] <skoolfile>\n' % sys.argv[0])
    sys.stderr.write('\nOptions:\n')
    sys.stderr.write("  -wX  Write only these elements, where X is one or more of:\n")
    sys.stderr.write("         %s = block types and addresses\n" % BLOCKS)
    sys.stderr.write("         %s = block titles\n" % BLOCK_TITLES)
    sys.stderr.write("         %s = block descriptions\n" % BLOCK_DESC)
    sys.stderr.write("         %s = registers\n" % REGISTERS)
    sys.stderr.write("         %s = mid-block comments and block end comments\n" % BLOCK_COMMENTS)
    sys.stderr.write("         %s = sub-block types and addresses\n" % SUBBLOCKS)
    sys.stderr.write("         %s = instruction-level comments\n" % COMMENTS)
    sys.stderr.write("  -h   Write addresses in hexadecimal format\n")
    sys.exit(1)

elements, WRITE_HEX = parse_options(sys.argv[1:-1])

# Create the parser
skoolfile = sys.argv[-1]
case = None
base = None
mode = 0
verbose = False
warn = False
fix_mode = 0
html = False
create_labels = False
mode_args = (case, base, mode, verbose, warn, fix_mode, html, create_labels)
parser = SkoolParser(skoolfile, mode_args)

# Write the ctl directives
write_ctls(parser.memory_map, elements)
